/*

    file:   video.c
    desc:   Video sync signal generator.

    note:   the algorthim is well optimized. I spent a lot of time searching the best way.

    author: Jaromir Dvorak (md@unicode.cz)

    This file is part of the AVGA engine.
 


    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
*/



#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include "video.h"
#include "driver.h"
#include "sound.h"


unsigned int c0;
unsigned char line, scroll;
volatile unsigned char wsync;
void *sp, *ss;


#ifdef VIDEO_USE_WINDOW
#include "../utils/window.h"
void *aw;
unsigned char height;
#else

unsigned char video_startline=0;
unsigned char video_params=0xF8;

void video_set_colormask(unsigned char mask)
{
	register unsigned char tmp = video_params;
	tmp&=~0xF0;
	mask&=0xF0;
	tmp|=mask;
	video_params = tmp;
}

static inline void video_set_scroll_x(unsigned char x)
{
	register unsigned char tmp = video_params;
	tmp&=~7;
	x&=7;
	tmp|=x;
	video_params = tmp;
}
static inline signed char video_rel_scroll_x(signed char x)
{
	register unsigned char xs = video_params;
	x+=xs&7;
	xs&=~7;
	xs|=x&7;
	video_params = xs;
	return x;
}
#if DRIVER_BLOCK_HEIGHT == 8
static inline signed char video_set_scroll_y(unsigned char yp)
{
	register unsigned char tmp = video_startline, y=yp;
	tmp&=~7; y&=7;	tmp|=y;
	video_startline = tmp;
}
static inline signed char video_rel_scroll_y(signed char yp)
{
	register unsigned char tmp = video_startline, y=yp;
	y+=tmp&7; tmp&=~7; tmp|=y&7;
	return y;
}
#else
static inline void video_set_scroll_y( unsigned char y)
{
	register unsigned char tmp = video_startline;
	tmp -= DRIVER_Y2PRE(tmp); 
	y = DRIVER_Y2PRE(y);
	tmp += y;
	video_startline = tmp;
}
static inline signed char video_rel_scroll_y(signed char y)
{
	register unsigned char tmp = video_startline;
	y += DRIVER_Y2PRE(tmp);
	tmp -= DRIVER_Y2PRE(tmp);
	tmp += DRIVER_Y2PRE(y);
	video_startline = tmp;
	return y;
}
#endif



#ifdef VIDEO_ABS_SCROLL
void video_set_scroll(unsigned char x, unsigned char y)
{
	video_set_scroll_x(x);
	video_set_scroll_y(y);
}
#endif

#ifdef VIDEO_REL_SCROLL
void video_scroll(signed char x, signed char y)
{
	video_rel_scroll_x(x);
	video_rel_scroll_y(y);
}	
#endif
#endif


void video_init(sync_t *s)
{
	sp = ss = (unsigned char*)s;
	c0 = 0;
	__video_timer_setup_mode14();
	__video_timer_select_OVF_IRQ();

	VIDEO_TIMER_OCRA = VIDEO_XPOS;
	VIDEO_TIMER_OCRB_DDR |= 1 << VIDEO_TIMER_OCRB_PIN;


#if defined(VIDEO_USE_WINDOW) || defined(WINDOW_RESET)
	window_reset();
#endif

}


unsigned int video_sync(void)
{
	unsigned int tick = 0;
	wsync = 0;
	while(!wsync) tick++;

	return tick;
}


ISR (VIDEO_TIMER_CMP_IRQ)
{
	if(scroll&0x08) 
	{
		VIDEO_TIMER_FINE_SYNC();
		driver_throw_scanline(line, scroll);
	}

#ifdef SOUND_LINEHANDLE
		sound_line_handle(); 					
#endif

	if(!--c0) //frame complete!
	{									
		wsync = 1;
#ifdef SOUND_FRAMEHANDLE
		sound_frame_handle();
#endif
		__video_timer_select_OVF_IRQ();
		__video_timer_clear_OVF_IRQ();
		return;
	}

#ifdef VIDEO_DOUBLE_SCANLINE
	if(c0&1) return; //only even lines
#endif

	line++;

#ifdef VIDEO_USE_WINDOW

	if(!--height) //window complete!
	{
		unsigned char *ptr = aw;
		register unsigned int tmp;

		line = *ptr++;
		height = *ptr++;
		scroll = *ptr++;


#ifdef WINDOW_INDIVIDUAL_TILESET
		asm volatile("ld %A0, %a1+ \r\n ld %B0, %a1+" : "=r" (tmp) : "e" (ptr) );
		_pgmp = tmp;
#endif
#ifdef WINDOW_INDIVIDUAL_REFTABLE
		asm volatile("ld %A0, %a1+ \r\n ld %B0, %a1+" : "=r" (tmp) : "e" (ptr) );
		_scrp = tmp;
#endif
		VIDEO_DDR = (scroll & 0xF0) | VIDEO_DEFAULT_DDR;
		aw = ptr;
	}
#else
	if(line == DRIVER_RESY) scroll = 0xF0;
#endif
}


ISR (VIDEO_TIMER_OVF_IRQ)
{
	static unsigned int top;

	VIDEO_TIMER_ICR = top;

	if(!c0)
	{
		register unsigned int tmp;
		register uint8_t* ptr = sp;

first_pulse:
		asm volatile("lpm %A0, %a1+ \r\n lpm %B0, %a1+" : "=r" (tmp) : "e" (ptr) );
		if(!tmp) { ptr = ss; goto first_pulse; }
		top = tmp;
		asm volatile("lpm %A0, %a1+ \r\n lpm %B0, %a1+" : "=r" (tmp) : "e" (ptr) );
		VIDEO_TIMER_OCRB = tmp;
		asm volatile("lpm %A0, %a1+ \r\n lpm %B0, %a1+" : "=r" (tmp) : "e" (ptr) );
		c0 = tmp;

		sp = ptr;
	}

	if(--c0 == VIDEO_YPOS)
	{

#ifdef VIDEO_USE_WINDOW
		register unsigned char* ptr = (unsigned char*)window_get_default();
		register unsigned int tmp;
		line = *ptr++;
		height = *ptr++;
		scroll = *ptr++;

#ifdef WINDOW_INDIVIDUAL_TILESET
		asm volatile("ld %A0, %a1+ \r\n ld %B0, %a1+" : "=r" (tmp) : "e" (ptr) );
		_pgmp = tmp;
#endif
#ifdef WINDOW_INDIVIDUAL_REFTABLE
		asm volatile("ld %A0, %a1+ \r\n ld %B0, %a1+" : "=r" (tmp) : "e" (ptr) );
		_scrp = tmp;
#endif
		aw = ptr;
#else
		line = video_startline;
		scroll = video_params;
#endif
		VIDEO_DDR = (scroll & 0xF0) | VIDEO_DEFAULT_DDR;
		

		__video_timer_select_CMP_IRQ();
		__video_timer_clear_CMP_IRQ();

	}

#ifdef SOUND_LINEHANDLE
		sound_line_handle(); 					
#endif
}












